/**
 * Copyright (C) 2000  Johan Redestig, Anders Palsson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.codemode.bishop;

import org.xml.sax.InputSource;
import org.xml.sax.HandlerBase;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.AttributeList;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.Parser;

import java.util.Hashtable;
import java.util.Stack;
import java.util.EmptyStackException;
import java.io.StringReader;
import java.io.IOException;
import java.io.DataInputStream;

import org.log4j.Category;

/**
 */
public class BishopParser extends HandlerBase
{
      private static final String DTD_FILE = "bishop_1.0.dtd";
      private Stack parseStack = new Stack();
      private Bishop bishop;
      private BishopEntity currentEntity;
      private String popName;
      private Hashtable classNames = new Hashtable();
      Category cat = Category.getInstance(Bishop.class.getName());
      protected Parser parser;
      
      private final static String[] classes = {"org.codemode.bishop.Bishop",
                                               "org.codemode.bishop.entity.Catch",
                                               "org.codemode.bishop.entity.Delete",
                                               "org.codemode.bishop.entity.Else",
                                               "org.codemode.bishop.entity.If",
                                               "org.codemode.bishop.Bishop",
                                               "org.codemode.bishop.entity.In",
                                               "org.codemode.bishop.entity.Invoke",
                                               "org.codemode.bishop.entity.Object",
                                               "org.codemode.bishop.entity.Out",
                                               "org.codemode.bishop.entity.Try",
                                               "org.codemode.bishop.entity.View",
                                               "org.codemode.bishop.entity.Trace"};
      
      private final static String[] tagNames = {"bishop",
                                                "catch",
                                                "delete",
                                                "else",
                                                "if",
                                                "import",
                                                "in",
                                                "invoke",
                                                "object",
                                                "out",
                                                "try",
                                                "view",
                                                "trace"};

      /* init the class names */
      {
         for (int index = 0; index < tagNames.length; index++)
         {
            classNames.put(tagNames[index].toUpperCase(), classes[index]);
         }
      }
      
      /**
       * Constructs a new ResponseHandler
       */
      public BishopParser() throws Exception
      {         
         SAXParserFactory parserFactory = SAXParserFactory.newInstance();
         
         parserFactory.setValidating(true);
         
         parser = parserFactory.newSAXParser().getParser();
         parser.setDocumentHandler(this);
         parser.setEntityResolver(this);
         parser.setErrorHandler(this);
      }
      
      /**
       * Parses the specifed xmlData, returns a Bishop
       */
      public Bishop parse(String xmlData) throws SAXException
      {
         try
         {
            bishop = new Bishop();
            currentEntity = null;
            popName = null;

            // the parser (startElement and endElement) sets the bishop
            parser.parse(new InputSource(new StringReader(xmlData)));
         }
         catch (Exception e)
         {
            throw new SAXException(e);
         }

         parseStack.clear();
         
         
         cat.debug("parsed complete, returns: " + bishop);         
         return bishop;
      }

      /**
       * Called when a new element is about to be parsed.
       */
      public void startElement(String tag, AttributeList attributes) throws SAXException
      {
         BishopEntity workObj;
         String classname;

         classname = (String)classNames.get(tag.toUpperCase());
         if (classname != null)
         {
            try
            {
               workObj = (BishopEntity)(Class.forName(classname).newInstance());
               workObj.setBishopTagName(tag);
               
               workObj.setAttributes(attributes);
               parseStack.push(workObj);
               popName = workObj.getBishopTagName();
            }
            catch(Exception e)
            {
               throw new SAXException(e);
            }
         }
         else
         {
            currentEntity = getCurrentEntity();
            popName = currentEntity.getBishopTagName();
         }
      }
      
      /**
       * Called when an element has been parsed.
       */
      public void endElement(String name) throws SAXException
      {
         BishopEntity workObj;
         BishopEntity peek;
         
         try
         {
            if (name.equals(popName))
            {
               workObj = (BishopEntity)parseStack.pop();
               
               if (parseStack.empty())
               {
                  bishop = (Bishop)workObj;
               }
               else
               {
                  peek = (BishopEntity)parseStack.peek();
                  peek.addChild(workObj);
                  workObj.setParent(peek);
                  popName = peek.getBishopTagName();
               }
               
               currentEntity = null;
            }
         }
         catch (Exception e)
         {
            throw new SAXException(e);
         }         
         currentEntity = null;
      }

      /**
       *
       */
      protected BishopEntity getCurrentEntity() throws SAXException
      {
         BishopEntity entity = null;
         
         try
         {
            if (parseStack.empty() == false)
            {
               entity = (BishopEntity)parseStack.peek();
            }
         }
         catch (Exception e)
         {
            throw new SAXException(e);
         }
         
         return entity;
      }

      /**
       * load dtd from classpath
       */
      public InputSource resolveEntity(String publicId, String systemId) throws SAXException    
      {
         // Create a InputSource with the cached dtd
         InputSource is = new InputSource(systemId);
         is.setByteStream(ClassLoader.getSystemResourceAsStream(DTD_FILE));
         is.setPublicId(publicId);
         return is;
      }      
      
      /**
       * Invoked by the parser when an error occurs
       */
      public void error (SAXParseException e) throws SAXException
      {
         throw new SAXException(e);
      }
      
      /**
       * Invoked by the parser when a warning occurs
       */
      public void warning (SAXParseException e) throws SAXException
      {
         throw new SAXException(e);
      }
      
      /**
       * Invoked by the parser when a fatal error occurs
       */
      public void fatalError(SAXParseException e)  throws SAXException
      {
         throw new SAXException(e);
      }
}
